<?php

	declare(strict_types=1);
		
/* PDO Database Class
 	http://wiki.hashphp.org/PDO_Tutorial_for_MySQL_Developers

	EXAMPLE
	
	require "shared/DB_PDO.php";

	$db = new DB;
	$db->connect($host, $database, $user, $password);


*/

	if (count(get_included_files()) < 2) { header("HTTP/1.1 301 Moved Permanently"); header("Location: /"); exit; } 


	class DB
	{
	
		public bool $hasConnection = false;
					
		private ?object $conn = null; // connection object
		
		public function drivers()
		{
			try
			{	
				return pdo_drivers();
						
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}				
		}
		
	
		public function connect(string $iHost, string $iDatabase, string $iUser, string $iPassword, ?string $iVendor = 'mysql', ?string $iCharset = 'utf8') : bool
		{// open database connection	
	
			try
			{								
				$ts1 = microtime(true);
		
				// connection string					
				$connStr = $iVendor . ':host=' . $iHost . ';dbname=' . $iDatabase . ';charset=' . $iCharset;
	
				// optons
				$options = array();
                
                	               									
				$result = false;
		
				// connect
				if($this->conn = new \PDO($connStr, $iUser, $iPassword, $options))
				{
					$result = true;
				}
				else
				{
					$ex = new Exception("Could not connect to database: " . $iDatabase);
				
					throw $ex;
					
					$result = false;
				}
								
				$ts2 = microtime(true);
							
				$this->hasConnection = $result;
				
				return $result;												
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}
	
				
		public function sql(string $iSQL, ?array $iParamsArr = null, ?bool $iURLencoded = false, ?bool $iNamedPlaceholders = false, ?bool $iFetch = false, ?bool $csvEscaping = false, ?string $csvSeparator = ',') // : array | bool
		{// read data from database
		
		/* 	$iSQL STRING
		 	$iParamsArr (ASSOC) ARRAY, evt. array in array for repetition
		 	$iURLencoded BOOLEAN
		 	$iNamedPlaceholder BOOLEAN 
			
			PDO::PARAM_NULL - 0
			PDO::PARAM_INT - 1
			PDO::PARAM_STR - 2
			PDO::PARAM_LOB - 3
			PDO::PARAM_BOOL - 5
			
		*/
		
		
			try 
			{		
								
			// query
									
				if($iParamsArr)
				{// binded

					// query
					$stmt = $this->conn->prepare($iSQL, array(\PDO::ATTR_CURSOR => \PDO::CURSOR_SCROLL)); // for iterate  , 
				
					if($iNamedPlaceholders == true)
					{// named placeholders
		
						if(gettype(array_key_first($iParamsArr)) == 'integer')
						{// array of arrays
				
							$arr = [];
				
							foreach($iParamsArr as $params)
							{
								$bindarr = [];
							
								foreach($params as $key => $value)
								{																		
									$bindarr[$key] = $value['value'];						
								}
								
								$arr[] = $bindarr;						
							}
														
							$stmt->execute($arr[0]); // NOT POSSIBLE TO BIND ARRAY OF NAMED PLACEHOLDERS 
									
						}
						else
						{

							foreach($iParamsArr as $key => $value)
							{															
								$stmt -> bindValue(':' . $key, $value['value'], $value['PDO']);
							}
							
							$stmt->execute();
						
						}
											
					}
					else
					{// not named placeholders (ie. by position)
						
						if(is_array($iParamsArr[0]))
						{// array of arrays
							foreach($iParamsArr as $params)
							{					
								// bind and execute
								$stmt->execute($params);	
							}
						}
						else
						{						
							// bind and execute
							$stmt->execute($iParamsArr);	
						}
					}
										
				}
				else
				{// Simple
			 
					// query		
					$stmt = $this -> conn -> query($iSQL, \PDO::FETCH_ASSOC);			 //
				}

				
				if($iFetch==true)
				{						
					return $this->fetchIterate($stmt, $iURLencoded, $csvEscaping, $csvSeparator); // array
				}
				else
				{
					$stmt -> closeCursor(); // this is not even required
					$stmt = null; // doing this is mandatory for connection to get closed
					unset($stmt);	
				}

			} 
			catch(\PDOException $e) 
			{		
								
				if(!is_null($this->conn))
				{
					try
					{				
						$this->conn->rollBack();					
					}
					catch(\PDOException $f)  
					{
					
					}					
				}
				
				throw $e;
							 
				return false;
			}			
		}
		
		private function fetch(object $stmt, ?bool $iURLencoded = false) : array
		{
			try
			{	
																					
				// fecth all
				$rows = $stmt -> fetchAll(\PDO::FETCH_ASSOC);
		//		$rows = $stmt->fetch();
				
				$stmt -> closeCursor();
						
				// new array
				$rowset = array();
		
				// looping thru rows			
				foreach($rows as $row)
				{
					$columnset = array(); // column collection
	
				//	$column_index = 0; // for meta

					// looping thru columns
					foreach($row as $name => $value)
					{	
													
						$columnArr = array('value'=>$value);
	
						// add url encoded value
						if($iURLencoded)
						{
							$columnArr['enc'] = urlencode($value);
						}
																	
						// set to column set
						$columnset[$name] = $columnArr;
						
					}	
					
					$rowset[] = $columnset;
												
				}
	
			//	$stmt -> closeCursor(); // this is not even required
				$stmt = null; // doing this is mandatory for connection to get closed
				unset($stmt);	
				
				return $rowset; // array			
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return [false];
			}						
		}
		
		
		private function fetchIterate(object $stmt, ?bool $iURLencoded = false) : array
		{ // https://www.dragonbe.com/2015/07/speeding-up-database-calls-with-pdo-and.html
		
			try
			{			
				// new array
				$rowset = array();
		
			
				while ($row = $stmt->fetch(\PDO::FETCH_ASSOC, \PDO::FETCH_ORI_NEXT))
				{
					$columnset = array(); // column collection
	
					// looping thru columns
					foreach($row as $name => $value)
					{	
												
						$columnArr = array('value'=>$value);
	
						// add url encoded value
						if($iURLencoded)
						{
							$columnArr['enc'] = urlencode($value);
						}
																						
						// set to column set
						$columnset[$name] = $columnArr;
												
					}	
					
					$rowset[] = $columnset;
												
				}
	
			
				$stmt -> closeCursor(); // this is not even required
			
				$stmt = null; // doing this is mandatory for connection to get closed
				unset($stmt);	
				
				return $rowset; // array			
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return [false];
			}			
			
		}
			
		public function begin() : bool
		{		
			try
			{						
				// begin transaction
				$this->conn->beginTransaction();
				
				return true;								
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}			
		}	
				
		public function commit() : bool
		{
			try
			{										
				// commit transaction
				$this->conn->commit();	
				
				return true;			
			} 
			catch(\PDOException $e) 
			{
				// roll back transaction				
				try
				{				
					$this->conn->rollBack();					
				}
				catch(\PDOException $f)  
				{
					
				}	
				
				throw $e;				
				
				return false;
													
			}		
		}	
		
		public function rollback()
		{		
			try
			{				
				$this->conn->rollBack();					
			}
			catch(\PDOException $f)  
			{
				
			}			
		}
			
		public function close() : bool
		{// close the database connection
		
			try
			{		
				
				$this->conn = null;	
				
				$this->hasConnection = false;
				
				return true;
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}
	
	
	
	// XML
	
		public function recordset2xml($iRecordset) : string | bool
		{// recordset		
			try
			{	
				if($this->log)
				{				
					$this->log->trace(__CLASS__, __FUNCTION__);
				}
					
				$xmlStr = "";
				$x = 0;
	
				foreach($iRecordset as $record)
				{		
					$x++;
						
					$xmlStr .= '<record no="' . $x . '">' ;
					
						$xmlStr .= $this->record2xml($record);

					$xmlStr .= '</record>';				
				}
	
				return $xmlStr;
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}
		
		private function record2xml($iRecord) : string | bool
		{// one record		
			try
			{						
				$xmlStr = "";
	
				foreach($iRecord as $column)
				{				
					$xmlStr .= '<col';
				
						foreach($column as $key => $value)
						{
							if($key!='value')
							{
								$xmlStr .= ' ' . $key . '="' . $value . '"';					
							}
						}
				
						$xmlStr .= '>' ;
			
						$xmlStr .= $this->_escape(strval($column['value']));

					$xmlStr .= '</col>' ;				
				}
	
				return $xmlStr;
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}				
		}
		
		public function recordset2xmlLitt(array $iRecordset, string $tablename) : string | bool
		{// recordset		
			try
			{
				
				$xmlStr = "";
				$x = 0;
	
				foreach($iRecordset as $record)
				{		
					$x++;
						
					$xmlStr .= '<' . $tablename . ' no="' . $x . '">' ;
					
						$xmlStr .= $this->record2xmlLitt($record); 

					$xmlStr .= '</' . $tablename . '>' ;				
				}
	
				return $xmlStr;
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}		
		
		public function record2xmlLitt($iRecord) : string | bool
		{// one record
		
			try
			{	
					
				$xmlStr = "";
	
				foreach($iRecord as $name => $column)
				{				
			
					$xmlStr .= '<' . $name;
				
						foreach($column as $key => $value)
						{
							if($key!='value')
							{
								$xmlStr .= ' ' . $key . '="' . $value . '"';					
							}
						}
				
						$xmlStr .= '>' ;
			
					
						$xmlStr .= strval($column['value']);					
			
					$xmlStr .= '</' . $name . '>' ;				
				}
	
				return $xmlStr;
			
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}
					
		private function _escape(string $iValue, ?bool $iDoubleEncode = false) : string | bool
		{// xml encode entities
	
			try
			{	
				
				if($iDoubleEncode == false)
				{
					$temp = $this->_unescape($iValue);
				}
				else
				{
					$temp = $iValue;	
				}
	
				$temp = str_replace('&','&amp;', $temp);  // &
	
				$temp = str_replace('"', '&quot;', $temp); // "
				$temp = str_replace('<', '&lt;', $temp);   // <
				$temp = str_replace('>', '&gt;', $temp);   // > 

				return $temp;
		
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}	
		}	
		
		private function _unescape(string $iValue) : string | bool
		{// xml decode entities
	
			try
			{				
				$temp = str_replace('&quot;', '"', $iValue); // "
				$temp = str_replace('&lt;', '<', $temp);   // <
				$temp = str_replace('&gt;', '>', $temp);   // > 
				$temp = str_replace('&apos;',"'", $temp);  // '
	
				$temp = str_replace('&amp;','&', $temp);  // &

				return $temp; 
	
			} 
			catch (\Throwable $e)
			{
				throw $e;				
				
				return false;
			}				
		}	
				
	} // class end

?>